#include <fstream.h>
#include <iostream.h>
#include <string.h>
#include <stdlib.h>
#include <iomanip.h>

#include "compareFiles.h"
#include "alignment.h"
#include "utils.h"

void menu(void);
void examples(void);

int main(int argc, char *argv[]){
  
  /* Parameters Control */
  bool appearErrors = false;
  float conserve = -1, gapThreshold = -1, simThreshold = -1, comThreshold = -1;
  int sgc = -1, sgt = -1, scc = -1, sct = -1, sfc = -1, sft = -1, windowSize = -1, compareset = -1, nogaps = -1, strict = -1, relaxed = -1;
  int stats = -1;

  /* Others varibles */
  ifstream compare;
  float *compareVect = NULL; 
  alignment **compAlig = NULL;
  similarityMatrix *similMatrix = NULL;
  int i = 1, lng, maxAminos = 0, numfiles = 0, referFile = 0;
  char *infile = NULL, *outfile = NULL, *matrix = NULL, **filesToCompare = NULL, line[256];

  /* Exec: trimAl - SHows the menu. */
  if(argc == 1) {
    menu(); return 0;
  }

  while(i < argc){

   /* Option -h ------------------------------------------------------------------------------------------- */
    if(utils::compare(argv[i], "-h") && (i+1 == argc)) {
      menu(); examples();
      return 0;
    }    
   /* Option -in ------------------------------------------------------------------------------------------- */
    if(utils::compare(argv[i], "-in") && (i+1 != argc) && (infile == NULL)) {

      if((sfc != -1) || (sft != -1 )) {
        cerr << endl << "ERROR: Not allowed combination with file compare statstics." << endl << endl;
        appearErrors = true;
        i++;
      }

      else if(compareset == -1) {
        lng = strlen(argv[++i]);
        infile = new char[lng + 1];      
        strcpy(infile, argv[i]);

        compAlig = new alignment*[1];
        compAlig[0] = new alignment;        

        if(!compAlig[referFile] -> loadAlignment(infile)) {
          cerr << endl << "ERROR: Alignment don't loaded: \"" << infile << "\" Check the file's content." << endl << endl;
          appearErrors = true;
        }
      }

      else {
        cerr << endl << "ERROR: Option \"" << argv[i] << "\" not valid. Exists a reference file with alignments to compare." << endl << endl;
        appearErrors = true;
        i++;
      } 
    }
   /* ------------------------------------------------------------------------------------------------------ */

   /* Option -out ------------------------------------------------------------------------------------------ */
    else if(utils::compare(argv[i], "-out") && (i+1 != argc) && (outfile == NULL)) {

      if(stats >= -1) {
        lng = strlen(argv[++i]);
        outfile = new char[lng + 1];
        strcpy(outfile, argv[i]);
        stats = i-1;
      }

      else {
        cerr << endl << "ERROR: Not allowed combinations with statistics options." << endl << endl;
        appearErrors = true;
        i++;
      }
    }
   /* ------------------------------------------------------------------------------------------------------ */

   /* Option -matrix --------------------------------------------------------------------------------------- */
    else if(utils::compare(argv[i], "-matrix") && (i+1 != argc) && (matrix == NULL)) {

      if((nogaps == -1) && (strict == -1) && (relaxed == -1)) {
        lng = strlen(argv[++i]);
        matrix = new char[lng + 1];
        strcpy(matrix, argv[i]);
      }

      else {
        cerr << endl << "ERROR: Not allowed combinations between automatics methods and similarity algorithm." << endl << endl;
        appearErrors = true;
        i++;
      }
    }
   /* ------------------------------------------------------------------------------------------------------ */

   /* Option -compareset ----------------------------------------------------------------------------------- */
    else if(utils::compare(argv[i], "-compareset") && (i+1 != argc) && (compareset == -1)) {

      if(infile == NULL) {
        compare.open(argv[++i], ifstream::in);
        if(!compare) {
          cerr << endl << "ERROR: Check the reference file with the alignments to compare." << endl << endl;
          appearErrors = true;
        }

        while(compare.getline(line, 256)) numfiles++;
        compare.close();

        compareset = i;
      }

      else {
        cerr << endl << "ERROR: Option \"" << argv[i] << "\" not valid. Exists a reference file with alignments to compare." << endl << endl;
        appearErrors = true;
        i++;
      }
    }
   /* ------------------------------------------------------------------------------------------------------ */

   /* Option -gt, gapthreshold ----------------------------------------------------------------------------- */
    else if((utils::compare(argv[i], "-gapthreshold") || utils::compare(argv[i], "-gt")) && (i+1 != argc) && (gapThreshold == -1)) {

      if(stats < -1) {
        cerr << endl << "ERROR: Not allowed combinations between " << argv[i] << " option and statistics options." << endl << endl;
        appearErrors = true;
        i++;
      }

      else if((nogaps != -1) || (strict != -1) || (relaxed != -1)) {
        cerr << endl << "ERROR: Not allowed combinations between automatics and manual methods." << endl << endl;
        appearErrors = true;
        i++;
      }

      else {
        if(utils::isNumber(argv[i+1])) {
          gapThreshold = 1 - atof(argv[++i]);
          if((gapThreshold < 0) || (gapThreshold > 1)) {
            cerr << endl << "ERROR: The gap threshold value must be between 0 and 1." << endl << endl;
            appearErrors = true;
            i++;
          }
          stats = i-1;
        }
        else {
          cerr << endl << "ERROR: The gap threshold value must be a real positive number." << endl << endl;
          appearErrors = true;
          i++;
        }
      }
    }
   /* ------------------------------------------------------------------------------------------------------ */

   /* Option -st -simthreshold ----------------------------------------------------------------------------- */
    else if((utils::compare(argv[i], "-simthreshold") || utils::compare(argv[i], "-st")) && (i+1 != argc) && (simThreshold == -1)) {

      if(stats < -1) {
        cerr << endl << "ERROR: Not allowed combinations between " << argv[i] << " option and statistics options." << endl << endl;
        appearErrors = true;
        i++;
      }

      else if((nogaps != -1) || (strict != -1) || (relaxed != -1)) {
        cerr << endl << "ERROR: Not allowed combinations between automatics and manual methods." << endl << endl;
        appearErrors = true;
        i++;
      }

      else {
        if(utils::isNumber(argv[i+1])) {
          simThreshold = atof(argv[++i]);
          if((simThreshold < 0) || (simThreshold > 1)) {
            cerr << endl << "ERROR: The similarity threshold value must be between 0 and 1." << endl << endl;
            appearErrors = true;
            i++;
          }
          stats = i-1;
        }
        else {
          cerr << endl << "ERROR: The similarity threshold value must be a real positive number." << endl << endl;
          appearErrors = true;
          i++;
        }
      }
    }
   /* ------------------------------------------------------------------------------------------------------ */

   /* Option -ct -comthreshold ----------------------------------------------------------------------------- */
    else if((utils::compare(argv[i], "-comthreshold") || utils::compare(argv[i], "-ct")) && (i+1 != argc) && (comThreshold == -1)) {

      if(stats < -1) {
        cerr << endl << "ERROR: Not allowed combinations between " << argv[i] << " option and statistics options." << endl << endl;
        appearErrors = true;
        i++;
      }

      else if((nogaps != -1) || (strict != -1) || (relaxed != -1)) {
        cerr << endl << "ERROR: Not allowed combinations between automatics and manual methods." << endl << endl;
        appearErrors = true;
        i++;
      }

      else {
        if(utils::isNumber(argv[i+1])) {
          comThreshold = atof(argv[++i]);
          if((comThreshold < 0) || (comThreshold > 1)) {
            cerr << endl << "ERROR: The compare threshold value must be between 0 and 1." << endl << endl;
            appearErrors = true;
            i++;
          }
          stats = i-1;
        }
        else {
          cerr << endl << "ERROR: The compare threshold value must be a real positive number." << endl << endl;
          appearErrors = true;
          i++;
        }
      }
    }
   /* ------------------------------------------------------------------------------------------------------ */

   /* Option -cons -conserve ------------------------------------------------------------------------------- */
    else if((utils::compare(argv[i], "-cons") || utils::compare(argv[i], "-conserve")) && (i+1 != argc) && (conserve == -1)) {

      if(stats < -1) {
        cerr << endl << "ERROR: Not allowed combinations between " << argv[i] << " option and statistics options." << endl << endl;
        appearErrors = true;
        i++;
      }

      else if((nogaps != -1) || (strict != -1) || (relaxed != -1)) {
        cerr << endl << "ERROR: Not allowed combinations between automatics and manual methods." << endl << endl;
        appearErrors = true;
        i++;
      }

      else {
        if(utils::isNumber(argv[i+1])) {
          conserve = atof(argv[++i]);
          if((conserve < 0) || (conserve > 100)) {
            cerr << endl << "ERROR: The minimal positions value must be between 0 and 100." << endl << endl;
            appearErrors = true;
            i++;
          }
          stats = i-1; 
       }
        else {
          cerr << endl << "ERROR: The minimal positions value must be a real/integer positive number." << endl << endl;
          appearErrors = true;
          i++;
        }
      }
    }
   /* ------------------------------------------------------------------------------------------------------ */

   /* Option -w -------------------------------------------------------------------------------------------- */
    else if(utils::compare(argv[i], "-w") && (i+1 != argc) && (windowSize == -1)){

      if((sfc != -1) || (sft != -1 )) {
        cerr << endl << "ERROR: Not allowed combination with file compare statstics." << endl << endl;
        appearErrors = true;
        i++;
      }

      else {
        if(utils::isNumber(argv[i+1])) {
          windowSize = atoi(argv[++i]);
          if(windowSize <= 0){
            cerr << endl << "ERROR: The window value must be a positive integer number." << endl << endl;
            appearErrors = true;
            i++;
          }
        }
        else {
          cerr << endl << "ERROR: The window value must be a number." << endl << endl;
          appearErrors = true;
          i++;
        }
      }
    }
   /* ------------------------------------------------------------------------------------------------------ */

   /* Option -nogaps --------------------------------------------------------------------------------------- */
    else if(utils::compare(argv[i], "-nogaps") && (nogaps == -1)) {

      if(stats < -1) {
        cerr << endl << "ERROR: Not allowed combinations between " << argv[i] << " option and statistics options." << endl << endl;
        appearErrors = true;
        i++;
      }

      else if((strict != -1) || (relaxed != -1)) {
        cerr << endl << "ERROR: Not allowed combinations between automatics methods." << endl << endl;
        appearErrors = true;
        i++;
      }

      else if((gapThreshold != -1) || (conserve != -1) || (simThreshold != -1)) {
        cerr << endl << "ERROR: Not allowed combinations between automatics and manual methods." << endl << endl;
        appearErrors = true;
        i++;
      }

      else {
        nogaps = 1;
        stats = i;
      }
    }
   /* ------------------------------------------------------------------------------------------------------ */

   /* Option -strict --------------------------------------------------------------------------------------- */
    else if(utils::compare(argv[i], "-strict") && (strict == -1)) {

      if(stats < -1) {
        cerr << endl << "ERROR: Not allowed combinations between " << argv[i] << " option and statistics options." << endl << endl;
        appearErrors = true;
        i++;
      }

      else if((nogaps != -1) || (relaxed != -1)) {
        cerr << endl << "ERROR: Not allowed combinations between automatics methods." << endl << endl;
        appearErrors = true;
        i++;
      }

      else if((gapThreshold != -1) || (conserve != -1) || (simThreshold != -1)) {
        cerr << endl << "ERROR: Not allowed combinations between automatics and manual methods." << endl << endl;
        appearErrors = true;
        i++;
      }

      else {
        strict = 1;
        stats = i;
      }
    }
   /* ------------------------------------------------------------------------------------------------------ */

   /* Option -relaxed -------------------------------------------------------------------------------------- */
    else if(utils::compare(argv[i], "-relaxed") && (relaxed == -1)) {

      if(stats < -1) {
        cerr << endl << "ERROR: Not allowed combinations between " << argv[i] << " option and statistics options." << endl << endl;
        appearErrors = true;
        i++;
      }

      else if((nogaps != -1) || (strict != -1)) {
        cerr << endl << "ERROR: Not allowed combinations between automatics methods." << endl << endl;
        appearErrors = true;
        i++;
      }

      else if((gapThreshold != -1) || (conserve != -1) || (simThreshold != -1)) {
        cerr << endl << "ERROR: Not allowed combinations between automatics and manual methods." << endl << endl;
        appearErrors = true;
        i++;
      }

      else {
        relaxed = 1;
        stats = i;
      }
    }
   /* ------------------------------------------------------------------------------------------------------ */
   /*  *********** ******************************** Statistics ********************************* *********** */

   /* Option -sgc ------------------------------------------------------------------------------------------ */
    else if((utils::compare(argv[i], "-sgc")) && (sgc == -1)) {
      if(stats < 0) {
        sgc = 1;
        stats--;
      }
      else {
        cerr << endl << "ERROR: Not allowed combination with " << argv[stats] << " option." << endl << endl;
        appearErrors = true;
        i++;
      }
    }
   /* ------------------------------------------------------------------------------------------------------ */

   /* Option -sgt ------------------------------------------------------------------------------------------ */
    else if((utils::compare(argv[i], "-sgt")) && (sgt == -1)) {      
      if(stats < 0) {
        sgt = 1;
        stats--;
      }
      else {
        cerr << endl << "ERROR: Not allowed combination with " << argv[stats] << " option." << endl << endl;
        appearErrors = true;
        i++;
      }
    }
   /* ------------------------------------------------------------------------------------------------------ */

   /* Option -scc ------------------------------------------------------------------------------------------ */
    else if((utils::compare(argv[i], "-scc")) && (scc == -1)) {
      if(stats < 0) {
        scc = 1;
        stats--;
      }
      else {
        cerr << endl << "ERROR: Not allowed combination with " << argv[stats] << " option." << endl << endl;
        appearErrors = true;
        i++;
      }
    }
   /* ------------------------------------------------------------------------------------------------------ */

   /* Option -sct ------------------------------------------------------------------------------------------ */
    else if((utils::compare(argv[i], "-sct")) && (sct == -1)) {
      if(stats < 0) {
        sct = 1;
        stats--;
      }
      else {
        cerr << endl << "ERROR: Not allowed combination with " << argv[stats] << " option." << endl << endl;
        appearErrors = true;
        i++;
      }
    }  
   /* ------------------------------------------------------------------------------------------------------ */

   /* Option -sfc ------------------------------------------------------------------------------------------ */
    else if((utils::compare(argv[i], "-sfc")) && (sfc == -1)) {

      if(windowSize != -1) {
        cerr << endl << "ERROR: Not allowed combination with -w option." << endl << endl;
        appearErrors = true;
        i++;
      }

      else if(infile != NULL) {
        cerr << endl << "ERROR: Not allowed combination with -in option." << endl << endl;
        appearErrors = true;
        i++;
      }

      else if(stats < 0) {
        sfc = 1;
        stats--;
      }
      else {
        cerr << endl << "ERROR: Not allowed combination with " << argv[stats] << " option." << endl << endl;
        appearErrors = true;
        i++;
      }
    }   
   /* ------------------------------------------------------------------------------------------------------ */

   /* Option -sft ------------------------------------------------------------------------------------------ */
    else if((utils::compare(argv[i], "-sft")) && (sft == -1)) {

      if(windowSize != -1) {
        cerr << endl << "ERROR: Not allowed combination with -w option." << endl << endl;
        appearErrors = true;
        i++;
      }

      else if(infile != NULL) {
        cerr << endl << "ERROR: Not allowed combination with -in option." << endl << endl;
        appearErrors = true;
        i++;
      }

      else if(stats < 0) {
        sft = 1;
        stats--;
      }

      else {
        cerr << endl << "ERROR: Not allowed combination with " << argv[stats] << " option." << endl << endl;
        appearErrors = true;
        i++;
      }
    } 
   /* ------------------------------------------------------------------------------------------------------ */

   /* Parameters not valids. ------------------------------------------------------------------------------- */
    else {
      cerr << endl << "ERROR: Parameter \"" << argv[i] << "\" not valid." << endl << endl;
      appearErrors = true;
    }    
   /* ------------------------------------------------------------------------------------------------------ */
    i++;
  }
    
  if(appearErrors) return -1;

  if(compareset != -1) {

    compAlig = new alignment*[numfiles];
    filesToCompare = new char*[numfiles];

    for(i = 0; i < numfiles; i++) 
    filesToCompare[i] = new char[256];

    compare.open(argv[compareset], ifstream::in);
    for(i = 0; i < numfiles; i++) {
      compare.getline(line, 256);

      compAlig[i] = new alignment;
      if(!compAlig[i] -> loadAlignment(line)) {
        cerr << endl << "Alignment don't loaded: \"" << line << "\" Check the file's content." << endl << endl;
        numfiles--;
        i--;
      }

      else {        
        strcpy(filesToCompare[i], line);
        compAlig[i] -> sequenMatrix();
        if(compAlig[i] -> getNumAminos() > maxAminos)
          maxAminos = compAlig[i] -> getNumAminos();      
      }
    } 
    compareVect = new float[maxAminos];

    if((gapThreshold == -1) && (conserve == -1) && (simThreshold == -1)) {

      if((comThreshold == -1) && (sft == -1) && (sfc == -1))
        referFile = compareFiles::algorithm(compAlig, filesToCompare, NULL, numfiles);

      else {
        referFile = compareFiles::algorithm(compAlig, filesToCompare, compareVect, numfiles);
        compAlig[referFile] = compAlig[referFile] -> cleanCompareFile(comThreshold, compareVect);
      }

    }
    else {

      if((comThreshold == -1) && (sft == -1) && (sfc == -1))
        referFile = compareFiles::algorithm(compAlig, numfiles, NULL);

      else {
        referFile = compareFiles::algorithm(compAlig, numfiles, compareVect);
        compAlig[referFile] = compAlig[referFile] -> cleanCompareFile(comThreshold, compareVect);       
      }

    }
  }
  
  if(gapThreshold == -1) gapThreshold = 0.0;
  if(conserve == -1)  conserve  = 100.0;
  if(simThreshold == -1) simThreshold = 100.0;

  if(windowSize != -1) {
    if(!compAlig[referFile] -> calculateGapStats(windowSize)) {
      cerr << endl << "ERROR: Troubles applying windowSize method." << endl << endl;
      return -1;
    }
  }

  if(nogaps != -1)
    compAlig[referFile] = compAlig[referFile] -> cleanGaps(0, 0);

  else if(strict != -1)
    compAlig[referFile] = compAlig[referFile] -> clean2ndSlope();

  else if(relaxed != -1)
    compAlig[referFile] = compAlig[referFile] -> cleanMixSlope();

  else if((simThreshold != 100.0) || (scc == 1) || (sct == 1)) {

    similMatrix = new similarityMatrix();
    if(matrix != NULL)
      similMatrix -> loadSimMatrix(matrix);
    else
      similMatrix -> loadSimMatrix("blosum62.mat");
 
    if(!compAlig[referFile] -> calculateConservationStats(similMatrix)) {     
      cerr << endl << "ERROR: Troubles proccesing the Similarity Matrix." << endl << endl;
      return -1;
    }
    
    if((gapThreshold != 0.0) || (conserve != 100.0))
      compAlig[referFile] = compAlig[referFile] -> clean(conserve, gapThreshold, simThreshold);

    else if(simThreshold != 100.0)
      compAlig[referFile] = compAlig[referFile] -> cleanConservation(conserve, simThreshold);
  }    

  else if((gapThreshold != 0.0) || (conserve != 100.0))
    compAlig[referFile] = compAlig[referFile] -> cleanGaps(conserve, gapThreshold);  

  if(outfile != NULL) {
    if(!compAlig[referFile] -> saveAlignment(outfile)) {
      cerr << endl << "ERROR: Troubles generating the output file." << endl << endl;
      return -1;
    }
  } 
  else if((sgc == -1) && (sgt == -1) && (scc == -1) && (sct == -1) && (sfc == -1) && (sft == -1))
    compAlig[referFile] -> printAlignment();
  
  if(sgc == 1) compAlig[referFile] -> printStatisticsGapsColumns();
  if(sgt == 1) compAlig[referFile] -> printStatisticsGapsTotal();
  if(scc == 1) compAlig[referFile] -> printStatisticsConservationColumns();
  if(sct == 1) compAlig[referFile] -> printStatisticsConservationTotal();

  if(compareset != -1) {
    if(sfc == 1) compareFiles::printStatisticsFileColumns(compAlig[referFile] -> getNumAminos(), compareVect);
    if(sft == 1) compareFiles::printStatisticsFileAcl(compAlig[referFile] -> getNumAminos(), compareVect);
  }

  return 0;
}

void menu(void) {

  cout << endl << endl;
  cout << "TrimAl v1.1 by Salvador Capella-Gutierrez, Jose M. Silla-Martínez and" << endl
       << "Toni Gabaldón." << endl << endl;

  cout << "http://bioinfo.cipf.es/tgabaldon/trimal" << endl
       << "This software is donated to the public domain." << endl << endl
       << "Please cite: Salvador Capella-Gutierrez, Jose M. Silla-Martínez and" << endl
       << "Toni Gabaldón. TrimAl: a tool for automatic alignment trimming" << endl
       << "(2008). (unpublished)." << endl << endl;

  cout << "Basic usage" << endl
       << "\ttrimal -in <inputfile> -out <outputfile> -(other options)." << endl << endl;

  cout << "Common options (for a complete list please see the User Guide or visit http://bioinfo.cipf.es/trimal):" << endl << endl;
  cout << "    -in <inputfile>          " << "Input file in PHYLIP format." << endl;
  cout << "    -out <outputfile>        " << "Output alignment in PHYLIP format (default stdout)." << endl;

  cout << "    -matrix <inpufile>       " << "Input file for user-defined similarity matrix (default is Blosum62)." << endl;
  cout << "    -compareset <inputfile>  " << "Input list of paths for the files containing the alignments to compare." << endl;

  cout << "    -gt -gapthreshold <n>    " << "1 - (fraction of sequences with a gap allowed)." << endl;
  cout << "    -st -simthreshold <n>    " << "Minimum average similarity allowed." << endl;

  cout << "    -ct -comthreshold <n>    " << "Minimum comparison value allowed." << endl;
  cout << "    -cons -conserve <n>      " << "Minimum percentage of the positions in the original alignment to conserve (overrides any threshold)." 
                                          << endl;

  cout << "    -w <n>                   " << "(half)Window size, score of position i is the average of the window (i - n) to (i + n)." << endl;
  cout << "    -nogaps                  " << "Remove all positions with gaps in the alignment (equivalent to -gt 0 -conserve 0)." << endl;

  cout << "    -strict                  " << "Use automatic selection on \"strict\" mode. (see User Guide)." << endl;
  cout << "    -relaxed                 " << "Use automatic selection on \"relaxed\" mode. (see User Guide)." << endl;

  cout << "    -sgc                     " << "Print gap percentage count for columns in the input alignment." << endl;
  cout << "    -sgt                     " << "Print accumulated gap percentage count." << endl;

  cout << "    -scc                     " << "Print conservation values for columns in the input alignment." << endl;
  cout << "    -sct                     " << "Print accumulated conservation values count." << endl;  

  cout << "    -sfc                     " << "Print compare values for columns in the selected alignment from compare files method." << endl;
  cout << "    -sft                     " << "Print accumulated compare values count for the selected alignment from compare files method."
                                          << endl << endl;  
}

void examples(void) {

  cout << "Some Examples:" << endl << endl;

  cout << "1) Removes all positions in the alignment with gaps in 10% or more of" << endl
       << "   the sequences, unless this leaves less than 60%. In such case, print" << endl
       << "   the 60% best (with less gaps) positions." << endl << endl;

  cout << "   trimal -in <inputfile> -out <outputfile> -gt 10 -cons 60" << endl << endl;

  cout << "2) As above but, the gap percentage is averaged over a window starting" << endl
       << "   3 positions before and ending 3 positions after the column." << endl << endl;

  cout << "   trimal -in <inputfile> -out <outputfile> -gt 10 -cons 60 -w 3" << endl << endl;

  cout << "3) Uses an automatic (strict) heuristic to decide optimal thresholds, based in the gap percentage" << endl
       << "   count over the whole alignment. (see User Guide for details)." << endl << endl;

  cout << "   trimal -in <inputfile> -out <outputfile> -strict" << endl << endl;
}

